<?php

/**
 * XoXo objects are generic and persistent objects. These objects will benefit
 * from the Oox_Registry and Xoxo_Storage implementation for all kind of object
 * manipulation.
 * 
 * They will inherit of a full administration interface, permissions handling
 * on full CRUD operations, and a database querying helpers.
 * 
 * The Xoxo_Storage implementation carries a full options CRUD handling and also
 * a strong and reliable caching mecanism. All load, update or insert operations
 * done on Xoxo_Object must always use the Xoxo_Storage methods in order for the
 * caching mecanism to fully work.
 */
class Xoxo_Object
  extends Oox_Registrable_Abstract
  implements Formable_Interface, Xoxo_Storable_Interface
{
  /**
   * Set current logged-in user as owner.
   */
  const OWNER_CURRENT = -1;

  /**
   * Current object storage, the one that allowed object load or save.
   * 
   * @var Xoxo_Storage
   */
// FIXME
//  private $__storage;

  /**
   * Get storage that manages this object.
   * 
   * @param Xoxo_Storage $storage
   * 
   * @return Xoxo_Object
   *   Self reference for chaining.
   */
// FIXME
/*
  public final function _setStorage(Xoxo_Storage $storage) {
    $this->__storage = $storage;
    return $this;
  }
 */

  /**
   * Return this particular object storage.
   * 
   * @return Xoxo_Storage
   *   Specific storage implementation.
   * 
   * @throws Xoxo_Storage_Exception
   *   If storage has not been set.
   */
// FIXME
/*
  public final function getStorage() {
    if (!isset($this->__storage)) {
      throw new Xoxo_Storage_Exception("Object has no storage set");
    }
    return $this->__storage;
  }
 */
  /**
   * Unique int identifier.
   * 
   * @var int
   */
  protected $_oid;

  /**
   * Set unique idendifier.
   * 
   * This is an internal method for storages, you must never set this manually,
   * like node nid, this is an automatic serial value from database.
   * 
   * @param int $oid
   * 
   * @return Xoxo_Object
   *   Self reference for chaining.
   */
  public function setIdentifier($oid) {
    $this->_oid = $oid;
    return $this;
  }

  /**
   * Get unique idendifier.
   * 
   * @return int
   * 
   * @throws XoxoException
   *   If the object does not exists in database.
   */
  public function getIdentifier() {
    if (!isset($this->_oid) || empty($this->_oid)) {
      throw new Oox_Exception("Identifier not set.");
    }
    return $this->_oid;
  }

  /**
   * Tell if the current object is stored.
   * 
   * @return boolean
   */
  public function isStored() {
    return isset($this->_oid);
  }

  /**
   * @var Schema_Interface
   */
  protected $_config;

  /**
   * Set human readable description.
   * 
   * @param string $description
   *   Human readable description.
   * 
   * @return Xoxo_Object
   *   Self reference for chaining.
   */
  public function setDescription($description) {
    $this->_config->set('description', $description);
    return $this;
  }

  /**
   * Get human readable description.
   * 
   * @return string
   */
  public function getDescription() {
    return $this->_config->get('description');
  }

  /**
   * Current status. 0 is disable, any other is enabled.
   * 
   * @var int
   */
  private $__status = Xoxo_Storable_Interface::STATUS_ENABLED;

  /**
   * Get object status.
   */
  public function getStatus() {
    return $this->__status;
  }

  /**
   * Set object status.
   * 
   * @param int $status
   *   New status.
   * 
   * @return Xoxo_Object
   *   Self reference for chaining.
   */
  public function setStatus($status) {
    $this->__status = (int) $status;
    return $this;
  }

  /**
   * Tells if object is enabled.
   * 
   * Alias to getStatus(), returns a boolean instead of an integer value.
   * 
   * @return boolean
   */
  public function isEnabled() {
    return $this->__status != Xoxo_Storable_Interface::STATUS_DISABLED;
  }

  /**
   * Object owner.
   * 
   * @var int
   *   Owner uid.
   */
  private $__uid = 0;

  /**
   * Get owner uid without trying to load it.
   *  
   * @return int
   *   Owner uid, can be 0 (anonymous).
   */
  public function getOwnerUid() {
    return $this->__uid;
  }

  /**
   * Get object owner.
   * 
   * @return object
   *   Drupal account object or FALSE if anonymous account.
   */
  public function getOwner() {
    if (is_numeric($this->__uid) && $account = user_load($this->__uid)) {
      return $account;
    }
    return FALSE;
  }

  /**
   * Set object owner.
   * 
   * @param int|object $account
   *   Owner uid or account object.
   *   It can also be the Xoxo_Object::OWNER_CURRENT constant for current
   *   user.
   * 
   * @return Xoxo_Object
   *   Self reference for chaining.
   */
  public function setOwner($account) {
    if (is_object($account)) {
      $account = $account->uid;
    }
    if ($account == Xoxo_Object::OWNER_CURRENT) {
      global $user;
      $this->__uid = $user->uid;
    }
    else {
      $this->__uid = is_object($account) ? $account->uid : $account;
    }
    return $this;
  }

  /**
   * Default constructor.
   */
  public function __construct() {
    global $user;
    $this->__uid = $user->uid;
    $this->_config = new Schema_Array(array(), FALSE, FALSE);
  }

  /**
   * Object just have been saved successfuly, it may have some custom Drupal
   * alterations to do after that.
   */
  public function postSave() {}

  /**
   * Render the preview page.
   */
  public function renderPreview() { return $this->getDescription(); }

  /**
   * Duplicate this instance without identifiers and unique keys.
   * 
   * @return Xoxo_Object
   */
  public function duplicate() {
    $object = clone $this;
    // Override some database and ownership related information. Most other
    // information will be gracefully duplicated using the clone statement.
    $object->setIdentifier(NULL);
    $object->setStatus(Xoxo_Storable_Interface::STATUS_ENABLED);
    $object->setOwner(Xoxo_Object::OWNER_CURRENT);
    $object->setDefault(FALSE);
    return $object;
  }

  /**
   * @see Formable_Interface::getForm()
   */
  public function getForm($formIdentifier = Formable_Interface::FORM_DEFAULT) {
    if (Formable_Interface::FORM_DEFAULT == $formIdentifier) {
      return new Xoxo_ObjectFormDefault;
    }
  }

  /**
   * @see Formable_Interface::getFormIdentifiers()
   */
  public function getFormIdentifiers() {
    return array(Formable_Interface::FORM_DEFAULT);
  }

  /**
   * @see Schema_Exportable_Interface::getSchema()
   */
  public function getSchema() {
    return clone $this->_config;
  }

  /**
   * Called at both load from database and restore from cache time.
   * 
   * @see Schema_Exportable_Interface::restoreFromSchema()
   */
  public function restoreFromSchema(Schema_Interface $config) {
    $this->_config->merge($config, TRUE);
    return $this;
  }
}
