<?php

/**
 * Defines basis for any kind of complex objets. 
 */
class Xoxo_Storage_Object_Exportable extends Xoxo_Storage_Object
{
  /**
   * @var array
   */
  protected $_loadedNames = array();

  /**
   * @var array
   */
  protected $_loadedNamesByOid = array();

  /**
   * Override the default constructor. This in order to save defaults as objects
   * in database for paging queries to work as any others.
   */
  public function _initStorage($item) {
    parent::_initStorage($item);
    // Check for defaults cache.
    if (!variable_get('xoxo_defaults_loaded', FALSE)) {
      // $this->saveDefaults();
    }
    return $this;
  }

  /**
   * Particular implementation that is able to use the object unique name as
   * caching parameter.
   * 
   * @see Xoxo_Storage::_cacheAdd()
   */
  public function _cacheAdd(Xoxo_Object $object, $toBinary = FALSE) {
    $oid = $object->getIdentifier();
    $name = $object->getName();

    // Parent cache (really important).
    $this->_loaded[$oid] = $object;
    // Our custom cache.
    $this->_loadedNames[$name] = $object;
    $this->_loadedNamesByOid[$oid] = $name;

    // Binary cache.
    if ($toBinary) {
      $clone = clone $object;
      // FIXME:: Ugly cache duplication. cache_get() function does not seems to
      // be able to fetch using wildcards.
      cache_set($this->getTableName() . '-' . $oid, $clone, XOXO_CACHE_OBJECT_TABLE);
      cache_set($this->getTableName() . '-' . $name, $clone, XOXO_CACHE_OBJECT_TABLE);
    }
  }

  /**
   * Particular implementation that is able to use the object unique name as
   * caching parameter.
   * 
   * @see Xoxo_Storage::_cacheGetBinary()
   */
  protected function _cacheGetBinary() {
    $args = func_get_args();

    $oid = isset($args[0]) ? $args[0] : NULL;
    $name = isset($args[1]) ? $args[1] : NULL;

    if ($oid) {
      $cid = $this->getTableName() . '-' . $oid;
    }
    else if ($name) {
      $cid = $this->getTableName() . '-' . $name;
    }

    if ($data = cache_get($cid, XOXO_CACHE_OBJECT_TABLE)) {
      // Proceed to some restoration.
      if ($data->data instanceof Xoxo_Object) {
// FIXME
//        $data->data->_setStorage($this);
      }
      return $data->data;
    }
  }

  /**
   * Get all defaults.
   */
  public function saveDefaults() {
    $hook = $this->_type . '_defaults';
    foreach (module_implements($hook) as $module) {
      foreach (module_invoke($module, $hook) as $object) {
        try {
          // For each default, check if an override exists, if not, save it into
          // database.
          if (!$this->existsByName($object->getName())) {
            // Explicitely set the default boolean.
            $object->setDefault(TRUE);
            // The save() method will update internal static cache, so there is
            // nothing else to care about here.
            $this->save($object);
          }
        }
        catch (Oox_Exception $e) {
          // Any exception here tells that the default definition is incomplete
          // so we just ignore it, but warn the developer he does an error.
          watchdog('xoxo', "A default object defined by module " . $module . " has a wrong definition (check if name is set), ignoring it.", NULL, WATCHDOG_ERROR);
        }
      }
    }
  }

  /**
   * @see Xoxo_Storage::_populate()
   */
  protected function _populate($row, Xoxo_Object $object) {
    parent::_populate($row, $object);
    if ($object instanceof Xoxo_Object_Exportable) {
      $object->setName($row->name);
    }
  }

  /**
   * Load layout by machine name.
   * 
   * @param string $name
   *   Machine name.
   * 
   * @return Xoxo_Object_Exportable
   *   Xoxo_Object_Exportable instance.
   * 
   * @throws Xoxo_Storage_Exception
   *   If object does not exists.
   */
  public function loadByName($name) {
    // Attempt cache fetch.
    if (!isset($this->_loadedNames[$name])) {
      $object = NULL;

      // Attempt from cache first time.
      if ($object = $this->_cacheGetBinary(NULL, $name)) {
        $this->_cacheAdd($object, FALSE);
      }

      // Load from database aptempt.
      else {
        $data = db_select($this->getTableName(), 'ct')
          ->fields('ct')
          ->condition('ct.name', $name)
          ->execute()
          ->fetchObject();

        if ($data) {
          // This will populate the new object.
          $object = $this->getItemRegistry()->getItem($data->type);
          $this->_populate($data, $object);
          $this->_cacheAdd($object, TRUE);
        }

        // Fallback, object does not exists.
        else {
          throw new Xoxo_Storage_Exception("Object '" . $name . "' does not exists.");
        }
      }
    }

    $ret = NULL;

    // Apply cloning method.
    switch ($this->_loadMethod) {
      case self::LOAD_METHOD_CLONE:
        $ret = clone($this->_loadedNames[$name]);
        break;
      case self::LOAD_METHOD_INSTANCE:
      default:
        $ret = $this->_loadedNames[$name];
    }

    return $ret;
  }

  /**
   * Tells if current layout name exists as a default layout.
   * 
   * @return boolean
   */
  public function existsAsDefault($name) {
    // FIXME: This is kind of ugly, and could be optimized.
    $hook = $this->_type . '_defaults';
    foreach (module_implements($hook) as $module) {
      foreach (module_invoke($module, $hook) as $object) {
        if ($object->getName() == $name) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

  /**
   * @see Xoxo_Storage::_preSave()
   * 
   * Last chance object save alteration, set our own real object name.
   */
  protected function _preSave(Xoxo_Object $object, &$data) {
    parent::_preSave($object, $data);
    $data['name'] = $object->getName();
  }

  /**
   * Delete given layout.
   * 
   * If layout does not exists, this method will remain silent.
   * 
   * @param string $name
   *   Xoxo_Object_Exportable name.
   */
  public function deleteByName($name) {
    return $this->deleteByField('name', $name);
  }

  /**
   * Check if given machine name already exists in database.
   * 
   * @param string $name
   *   Machine name.
   * 
   * @return boolean
   */
  public function existsByName($name) {
    return (bool) db_query("SELECT 1 FROM {:table} WHERE name = :name", array(
      ':table' => $this->getTableName(),
      ':name' => $name
    ))->fetchField();
  }

  /**
   * Import layout from raw PHP code.
   * In case of any error, no records at all will be saved.
   * 
   * Be careful, if any of these layouts exists, it will overwrite the
   * existing one.
   * 
   * Note for developers: when you override this method, be careful about the
   * $pretend parameter, and dont forget to call the Xoxo_Storage::save()
   * method in order to save new objects.
   * 
   * @param string $code
   *   Raw PHP code returning a Xoxo_Object_Exportable instance, or an array of
   *   Xoxo_Object_Exportable instances.
   * @param boolean $pretend = FALSE
   *   Set this to TRUE if you don't want code to be actually imported. This
   *   can be usefull for forms validation phase.
   * 
   * @throws Xoxo_Storage_Exception
   *   In case of any error. Error message will be a localized human readable
   *   error message and will be displayed to end use in the import form.
   * 
   * @return array
   *   The array of new imported layouts.
   */
  public function importCode($code, $pretend = FALSE) {
    throw new Xoxo_Storage_Exception("Not implemented");
  }

  /**
   * @see Xoxo_Storage::_cacheClear()
   */
  protected function _cacheClear($oids = array()) {
    parent::_cacheClear($oids);
    // An optimized implementation would do it in a single foreach() by fully
    // overriding the parent method. Still keeping it like this for other
    // implementors to have a sample of how to do it without breaking parent
    // implementation.
    if (!empty($oids)) {
      foreach ($oids as $oid) {
        $name = $this->_loadedNamesByOid[$oid];
        unset($this->_loadedNames[$name], $this->_loadedNamesByOid[$oid]);

        // Remove binary cached object on a per name basis.
        cache_clear_all($this->getTableName() . '-' . $name, XOXO_CACHE_OBJECT_TABLE);
      }
    }
    else {
      $this->_loadedNames = array();
      $this->_loadedNamesByOid = array();
    }
  }
}
