<?php

/**
 * Registry implementation for storages. We are going to override the
 * getItem() method in order to maintain a pool of singletons instead
 * of creating new instances each time.
 */
class Xoxo_Storage_Registry extends Oox_Registry
{
  /**
   * @see Oox_Registry::_checkItem()
   */
  protected function _checkItem($type, &$item, $module) {
    // Check for registry type.
    $registry = NULL;
    $registry_type = isset($item['registry_type']) ? $item['registry_type'] : $type;
    try {
      $registry = Oox_Registry::getInstance($registry_type);
    }
    catch (Oox_Registry_Exception $e) {
      watchdog('xoxo', "Module " . $module . " is missing a registry type, guessed or given type is: " . $registry_type, NULL, WATCHDOG_ERROR);
      return FALSE;
    }
    // If no class is provided, give the one with the least features we can to
    // ensure it will fit with objects the user would want to save.
    if (!isset($item['class'])) {
      $item['class'] = 'Xoxo_Storage';
      watchdog('xoxo', "Module ' . $module . ' defines the storage type '" . $registry_type . "' which provies no class, set default to 'Xoxo_Storage'", NULL, WATCHDOG_ALERT);
    }
    // Check path.
    if (isset($item['path'])) {
      // FIXME: Do some cleaning here.
      $item['path'] = $item['path'];
    }
    else {
      // This is the default path, see hook_menu() implementation.
      $item['path'] = 'admin/objects/' . $type;
    }
    // This test must be done after we may have set a default class name.
    if (parent::_checkItem($type, $item, $module)) {
      // Check for administrable ability.
      if (!isset($item['administrable'])) {
        $item['administrable'] = FALSE;
      }
      // Check for load method.
      if (!isset($item['load_method'])) {
        $item['load_method'] = Xoxo_Storage::LOAD_METHOD_INSTANCE;
      }
      // We are all good, I presume.
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Singleton pattern override.
   * 
   * @var array
   */
  private $__instances = array();

  /**
   * Overrided in order to keep the old hook name.
   * 
   * @see Oox_Registry::_getHookName()
   */
  public function _getHookName() {
    return 'xoxo_storage';
  }

  /**
   * getItem() method override, we are here going to implement a singleton
   * pattern for the current registry elements and ensure storage.
   * 
   * @see Oox_Registry::getItem()
   */
  public function getItem($type) {
    $this->_loadItemCache();

    // Create new instance and return element if all ok.
    if (!isset($this->__instances[$type])) {
      $this->__instances[$type] = parent::getItem($type)->_initStorage($this->_cache[$type]);
    }

    // Return the object.
    return $this->__instances[$type];
  }
}
