<?php

/**
 * @file
 * XoXo Admin UI pages.
 * 
 * Notice that in all form functions we will prefer to use storage cached
 * description rather than the object implemented interfaces. We do this
 * because in those descriptions, the developper could had disabled some
 * features support to tune these forms, but keeping the interfaces
 * implementation in order to pragmatically use them in his code.
 * 
 * This will be the one and only case we use those description, in the OOP
 * API library, we will rely on objects interfaces implementations instead,
 * to ensure the best behavior.
 */

/**
 * Default item count for administration paged tables.
 */
define('XOXO_OBJECT_ADMIN_LIMIT', 16);

/**
 * Helper to build some url.
 * 
 * @param XoxoObject $object
 * 
 * @return int|string
 *   Object oid if not exportable, else object unique name.
 */
function _xoxo_storage_admin_get_identifier($object) {
  if ($object instanceof Xoxo_Object_Exportable) {
    try {
      return $object->getName();
    }
    catch (OoxException $e) { }
  }
  else if ($object instanceof Xoxo_Storable_Interface) {
    try {
      return $object->getIdentifier();
    }
    catch (Oox_Exception $e) { }
  }
  return FALSE;
}

/**
 * Ajax callback for overview form that toggles an object status.
 */
function xoxo_storage_ajax_toggle_status($type, $object) {
  try {
    switch ($object->getStatus()) {
      case Xoxo_Object::STATUS_DISABLED:
        $status = Xoxo_Object::STATUS_ENABLED;
        break;
      case Xoxo_Object::STATUS_ENABLED:
      default:
        $status = Xoxo_Object::STATUS_DISABLED;
        break;
    }
    $object->setStatus($status);
    xoxo_storage_get($type)->saveStatus($object);
    print json_encode(array('success' => TRUE, 'status' => $status));
  }
  catch (Oox_Exception $e) {
    print json_encode(array('success' => FALSE));
  }
  exit();
}

/**
 * Storage headers.
 */
function xoxo_storage_global_overview_headers() {
  return array(
    array('data' => t("Enabled"), 'field' => 'ct.status'),
    array('data' => t("Identifier"), 'field' => 'ct.oid'),
    array('data' => t("Description")),
    array('data' => t('Owner'), 'field' => 'ct.uid'),
    array('data' => t("Type"), 'field' => 'ct.type'),
    array('data' => t("Operations")),
  );
}

/**
 * Global overview form, can display either a named objects listing, or an
 * integer identified objects listing.
 * 
 * FIXME: Function parameters.
 */
function xoxo_storage_global_overview_form($form, &$form_state, $type) {
  $form = array();

  // Set some variable we could want to get back in validation and submission form.
  $form['#object_type'] = $type;

  // Load associated object storage.
  $storage = xoxo_storage_get($type);

  // Allow custom storages to set an help message in this form.
  $form['help'] = array(
    '#markup' => $storage->getOverviewHelp(),
    '#prefix' => '<p>',
    '#suffix' => '</p>',
  );

  // Path prefix for specific storage administration screen.
  $storage_path = $storage->getPath();

  // Create an array of loaded objects.
  $form['objects'] = array();
  
  // Build a clean query.
  $oids = $storage
    ->getQuery('ct', 'view', FALSE)
    ->extend('PagerDefault')
    ->extend('TableSort')
    ->limit(XOXO_OBJECT_ADMIN_LIMIT)
    ->orderByHeader(xoxo_storage_global_overview_headers())
    ->element(0)
    ->execute()
    ->fetchCol();

  if (!empty($oids)) {
    $registry = $storage->getItemRegistry();
    foreach ($storage->load($oids) as $object) {
      // Get the identifier, it can be different weither the object is exportble
      // or not. This identifier will be used for all link path. The menu routeur
      // load function is able to loaded by oid or name indifferently so this is
      // only cosmetics.
      $identifier = _xoxo_storage_admin_get_identifier($object);
      $form['objects'][$identifier] = array();
  
      // Set enabled checkbox, this will be override by theme function in order to
      // nicely use ajax with it.
      $enabled = $object->isEnabled();
      $form['objects'][$identifier]['enabled'] = array(
        '#type' => 'checkbox',
        '#attributes' => array('class' => array('xoxo-enabled')), 
        '#default_value' => $enabled,
      );
      $form['objects'][$identifier]['identifier'] = array('#markup' => $identifier);
  
      // Compute description field, with preview link.
      $description = $object->getDescription();
      if (!$description) {
        $description = t("Overview <em>(no description available)</em>");
      }
      $description = l($description, $storage_path . '/' . $identifier, array('html' => TRUE, 'attributes' => array('title' => t("Preview"))));
      $form['objects'][$identifier]['description'] = array('#markup' => $description);

      // Set owner link.
      $form['objects'][$identifier]['owner'] = array('#markup' => theme('username', array('account' => $object->getOwner())));

      // Type.
      $form['objects'][$identifier]['registry_type'] = array('#markup' => $registry->getItemName($object->getType()));

      // Operations.
      $links = array();
      // FIXME
      /*
      if ($object->isDefault()) {
        $links[] = array(
          'href' => $storage_path . '/' . $identifier . '/edit',
          'title' => t('Override'),
        );
      }
      else {
       */
        $links[] = array(
          'href' => $storage_path . '/' . $identifier . '/edit',
          'title' => t('Edit'),
        );
        $links[] = array(
          'href' => $storage_path . '/' . $identifier . '/delete',
          'title' => t('Delete'),
        );
      /*
      }
       */
      if (oox_object_type_is_exportable($type)) {
        $links[] = array(
          'href' => $storage_path . '/' . $identifier . '/export',
          'title' => t('Export'),
        );
      }
      $form['objects'][$identifier]['operations'] = array('#markup' => theme('links', array('links' => $links)));
    }
  }

  $form['pager'] = array('#markup' => theme('pager', array()));
  return $form;
}

/**
 * Import form, type dependant.
 */
function xoxo_storage_global_import_form($form, &$form_state, $type) {
  $form = array();
  $form['#object_type'] = $type;
  $form['import'] = array(
    '#type' => 'textarea',
    '#title' => t("Exported PHP code"),
    '#rows' => 10,
    '#description' => t("Cut and paste exported code here to import in this site."),
  );
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t("Import"),
  );
  return $form;
}

/**
 * Import form validate handler.
 */
function xoxo_storage_global_import_form_validate($form, &$form_state) {
  try {
    xoxo_storage_get($form['#object_type'])->importCode($form_state['values']['import'], TRUE);
  }
  catch (Exception $e) {
    form_set_error('import', $e->getMessage());
  }
}

/**
 * Import form submit handler.
 */
function xoxo_storage_global_import_form_submit($form, &$form_state) {
  $storage = xoxo_storage_get($form['#object_type']);
  $objects = $storage->importCode($form_state['values']['import']);
  foreach ($objects as $object) {
    drupal_set_message(t("<strong>@object</strong> has been saved.", array('@object' => $object->getName())));
  }
  if (count($objects) == 1) {
    drupal_goto($storage->getPath() . '/' . array_shift($objects)->getName());
  }
  else {
    drupal_goto($storage->getPath());
  }
}

/**
 * Add an object page.
 */
function xoxo_storage_object_add_form($form, &$form_state, $type) {
  $storage = xoxo_storage_get($type);
  if ($properties = $storage->getItemRegistryProperties()) {
    $types = array_keys($storage->getItemRegistry()->getItemsByProperties($properties));
  }
  else {
    $types = array_keys($storage->getItemRegistry()->getAllItems());
  }

  // Skip the selection form if only one object type is available.
  if (count($types) == 1) {
    $form_state['values']['item_type'] = $types[0];
  }

  // Hazardous implementation of multistep form.
  if (isset($form_state['values']['item_type'])) {
    $form += xoxo_storage_object_edit_form($form, &$form_state, $type, $form_state['values']['item_type']);
  }
  else {
    $form += xoxo_storage_object_select_type_form($form, &$form_state, $type);
  }

  return $form;
}

/**
 * Pre-add form, ensure we have an object type.
 */
function xoxo_storage_object_select_type_form($form, &$form_state, $type) {
  $form['#tree'] = TRUE;
  $storage = xoxo_storage_get($type);

  $form['item_type'] = array(
    '#type' => 'oox_registry_item',
    '#title' => t("Select object type"),
    // '#options' => $options,
    '#registry_type' => $storage->getItemRegistry()->getType(),
    '#default_value' => NULL,
    '#required' => TRUE,
    '#properties' => $storage->getItemRegistryProperties(),
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t("Use this type"),
    '#validate' => array(),
    '#submit' => array('xoxo_storage_object_select_type_form_submit'),
  );
  return $form;
}

/**
 * Object type select submit handler.
 */
function xoxo_storage_object_select_type_form_submit($form, &$form_state) {
  $form_state['rebuild'] = TRUE;
}

/**
 * Add, Edit, Override form.
 */
function xoxo_storage_object_edit_form($form, &$form_state, $type, $object) {
  $form['#tree'] = TRUE;

  $storage = xoxo_storage_get($type);

  // Prepare some variables.
  $form['#object_type'] = $type;
  $form['#object_create'] = TRUE;

  if ($object instanceof Xoxo_Storable_Interface) {
    try {
      $identifier = _xoxo_storage_admin_get_identifier($object);
      // FIXME
      /*
      if (!$object->isDefault()) {
       */
        $label = t("Update");
      /*
      }
      else {
        $label = t("Override");
      }
       */
      $form['#object_create'] = FALSE;
      $form_state['object'] = $object;
    }
    catch (Exception $e) {
      $object = $form_state['object'] = $storage->getItemRegistry()->getItem($object);
      $label = t('Add');
    }
  }
  else {
    $object = $form_state['object'] = $storage->getItemRegistry()->getItem($object);
    $label = t('Add');
  }

  // Specific object subform.
  $form['object_form'] = array(
    '#type' => 'oox_formable',
    '#object' => $object,
  );

  $form['buttons'] = array(
    '#prefix' => '<div class="container-inline">',
    '#suffix' => '</div><div class="clear-block"></div>',
  );
  $form['buttons']['submit'] = array(
    '#type' => 'submit',
    '#value' => $label,
    '#submit' => array('xoxo_storage_object_edit_form_submit'),
  );
  if (!$form['#object_create']) {
    $form['buttons']['delete'] = array(
      '#type' => 'submit',
      '#value' => t("Delete"),
      '#validate' => array(),
      '#submit' => array('xoxo_storage_object_edit_form_submit_delete'),
    );
  }

  if (isset($_GET['destination']) || isset($form_state['redirect'])) {
    // Choose the right path.
    $path = isset($_GET['destination']) ? $_GET['destination'] : $form_state['redirect'];
    $form['buttons']['cancel'] = array('#markup' => l(t("Cancel"), $path));
  }

  return $form;
}

/**
 * Edit form submit handler.
 */
function xoxo_storage_object_edit_form_submit($form, &$form_state) {
  $hasError = FALSE;
  $storage_path = NULL;

  try {
    // Fetch storage before anything else.
    $storage = xoxo_storage_get($form['#object_type']);
    $storage_path = $storage->getPath();

    // Load instance and fill it with form values.
    $object = $form_state['values']['object_form'];

    // Save.
    $ret = $storage->save($object);
    $identifier = _xoxo_storage_admin_get_identifier($object);

    // In order to be the more accurate possible, we display the status message
    // following the save() method return, not using the form parameters, that
    // could be asynchronized will real object state.
    switch ($ret) {
      case Xoxo_Storage::SAVED_NEW:
        drupal_set_message(t("Object @object has been added.", array('@object' => $identifier)));
        break;
      case Xoxo_Storage::SAVED_UPDATED:
        drupal_set_message(t("Object @object has been updated.", array('@object' => $identifier)));
        break;
      default:
      case Xoxo_Storage::SAVED_ERROR:
        drupal_set_message(t("Object @object could not be saved.", array('@object' => $identifier)), 'error');
        $hasError = TRUE;
        break;
    }
  }
  catch (Exception $e) {
    $hasError = TRUE;
    watchdog('xoxo_storage', $e);
    drupal_set_message($e->getMessage(), 'error');
  }

  // Allow caller to get back the new object identifier as GET parameter.
  // But we will do this only in case all went ok.
  if (!$hasError) {
    $form_state['redirect'] = $storage_path ? $storage_path .'/' . $identifier : 'admin/objects';
  }
}

/**
 * Delete button submit handler.
 */
function xoxo_storage_object_edit_form_submit_delete($form, &$form_state) {
  // TODO
  // $form_state['redirect'] = $storage->getPath() . '/' . $oid . '/delete';
}

/**
 * Export object form.
 */
function xoxo_storage_object_export_form($form, &$form_state, $type, $object) {
  $form['#attached'] = array('js' => array(
    drupal_get_path('module', 'oox') . '/js/jquery.copy.min.js',
    drupal_get_path('module', 'oox') . '/js/copy.js',
  ));
  $form['object_type'] = array(
    '#type' => 'value',
    '#value' => $type,
  );

  // Check if we have a value for export.
  $format = isset($_GET['format']) ? $_GET['format'] : 'json';

  $form['format'] = array(
    '#prefix' => '<div class="container-inline">',
    '#suffix' => '</div>',
  );
  $form['format']['format'] = array(
    '#type' => 'select',
    '#title' => t("Format"),
    '#options' => array(
      'json' => t("JSON"),
    ),
    '#default_value' => $format,
  );
  $form['format']['submit'] = array(
    '#type' => 'submit',
    '#value' => t("Export"),
    '#submit' => array('xoxo_storage_object_export_form_change_format_submit'),
  );

  switch ($format) {
    case 'json':
      $writer = new Schema_Writer_Json($object->getSchema());
      break;
  }

  if (isset($writer)) {
    $form['export'] = array(
      '#type' => 'markup',
      '#prefix' => '<pre id="xoxo-export-text" style="background: white; color: black; padding: 0.5em 1em; border: 1px solid black;">',
      '#suffix' => '</pre>',
      '#markup' => (string)$writer,
    );
    $form['clip'] = array(
      '#type' => 'button',
      '#value' => t("Copy to clipboard"),
      '#attributes' => array('class' => array('copy-to-clipboard'), 'rel' => 'xoxo-export-text'),
    );
  //  $form['exporta'] = array(
  //    '#type' => 'textarea',
  //    '#title' => t("Exported PHP code"),
  //    '#rows' => 10,
  //    '#description' => t("Cut and paste this code into any other Drupal installation import form, or in any custom module code."),
  //    '#default_value' => implode("\n", xoxo_storage_get($type)->export($object)),
  //  );
  }
  return $form;
}

/**
 * Change format submit handler for export form.
 */
function xoxo_storage_object_export_form_change_format_submit($form, &$form_state) {
  $form_state['redirect'] = array($_GET['q'], array('query' => array('format' => $form_state['values']['format'])));
}

/**
 * Delete object confirmation form.
 */
function xoxo_storage_object_delete_form($form, &$form_state, $type, $object) {
  $form = array('oid' => array('#type' => 'value', '#value' => $object->getIdentifier()));
  $form['#object_type'] = $type;
  $storage = xoxo_storage_get($type);
  $description = NULL;
  // Try to instrospect for a default.
  if ($storage instanceof Xoxo_Storage_Object_Exportable) {
    $name = $object->getName();
    /*
    if ($storage->existsAsDefault($name)) {
      $description = t("This object will be reverted to default, any manual modifications will be loosed.");
    }
     */
    $question = t("Are you sure you want to delete the <em>@object</em> object?", array(
      '@object' => $name,
    ));
  }
  else {
    $question = t("Are you sure you want to delete this object ?");
  }
  if (!$description) {
    $description = t("This object will be definitively deleted.");
  }
  return confirm_form($form, $question, $storage->getPath() . '/overview', $description);
}

/**
 * Delete object confirmation form submit handler.
 */
function xoxo_storage_object_delete_form_submit($form, &$form_state) {
  $storage = xoxo_storage_get($form['#object_type']);
  $storage->deleteByIdentifier($form_state['values']['oid']);
  $form_state['redirect'] = $storage->getPath() . '/overview';
}

/**
 * Object preview page.
 */
function xoxo_storage_object_preview_page($type, $object) {
  return theme('xoxo_object', array('object' => $object, 'preview' => TRUE));
}

/**
 * Submit handler for library form clear cache button.
 */
function xoxo_registry_manage_submit_clear_cache($form, &$form_state) {
  cache_clear_all('oox_library_path', OOX_CACHE_TABLE);
  drupal_set_message(t("Library cache has been cleared, cache will be completed each time a contribution module asks for an external library."));
}


/**
 * Theme function for global overview form.
 */
function theme_xoxo_storage_global_overview_form(&$variables) {
  $form = &$variables['form'];

  // If we got here, storage is available.
  $storage = xoxo_storage_get($form['#object_type']);

  // Provide UI alteration that will enable ajax nice stuff.
  drupal_add_js(drupal_get_path('module', 'xoxo') . '/xoxo.admin.js');
  drupal_add_js(array('Xoxo' => array(
    'ajaxPrefixUrl' => url($storage->getPath() . '/ajax'),
  )), 'setting');
  drupal_add_css(drupal_get_path('module', 'xoxo') . '/xoxo.admin.css');

  // Prepare table rendering.
  $headers = xoxo_storage_global_overview_headers();
  $rows = array();

  foreach (element_children($form['objects']) as $key) {
    // Extract some useful vars.
    $enabled = (bool) $form['objects'][$key]['enabled']['#default_value'];
    $name = $form['objects'][$key]['identifier']['#markup'];

    // Prepare current row.
    $row = array(
      'data' => array(),
      // Helpers for css/js/ajax handling.
      'class' => array('xoxo-object-row', ($enabled ? 'xoxo-object-row-enabled' : 'xoxo-object-row-disabled')),
      'rel' => $name,
    );

    // Set some already computed cells.
    $row['data'][] = drupal_render($form['objects'][$key]['enabled']);
    $row['data'][] = drupal_render($form['objects'][$key]['identifier']);
    $row['data'][] = drupal_render($form['objects'][$key]['description']);
    $row['data'][] = drupal_render($form['objects'][$key]['owner']);
    $row['data'][] = drupal_render($form['objects'][$key]['registry_type']);
    $row['data'][] = drupal_render($form['objects'][$key]['operations']);

    $rows[] = $row;
  }

  return theme('table', array('header' => $headers, 'rows' => $rows)) . drupal_render_children($form);
}
