<?php
// $Id: xoxo.api.php,v 1.1.4.4 2011/01/04 18:25:40 pounard Exp $

/**
 * @file
 * XoXo module hooks definition and documentation.
 */

/**
 * This hook allows modules to use the default oox module administration
 * UI to manage their exportables.
 * 
 * Implementing this hook will implicitely create a new hook called after your 
 * object type: hook_OBJECT_TYPE_defaults.
 * 
 * Implementing this hook using a custom storage will also implicitely create
 * the features module oox to allow you to import/export your modules.
 * In order for this to work, this is mandatory that you override the import
 * export methods in your custom storage.
 * 
 * A storage is tied to an existing registry, any objects from this registry
 * will be CRUD-able using the defined storage.
 * You can restrict object types you would want to support.
 * 
 * @return array
 *   It must return a valid array, following this convention:
 *     - Each key is an exportable type name. This is the name you will get in
 *       the menu pathes. Value is an array filled with options.
 *   Following are array options:
 *     - 'name' : string
 *       Human readable type name, not localized.
 *     - 'administrable' : boolean
 *       (optional) default is FALSE
 *       TRUE or FALSE, will tell the oox module to build the full
 *       administration tools to manage object instances.
 *     - 'table' : string
 *       (required)
 *       This setting will tell the table name where exportable objets are
 *       stored. Note that this table must have a compatible schema with
 *       the 'oox_data' table.
 *       Whatever happens, oox module will try to create this table
 *       using the right schema.
 *     - 'registry_type' : string
 *       (optional) Specify here the registry type you want to use. The registry
 *       type must exists and be defined by the hook_oox_registry().
 *       The registry will spawn new object instances on load, before populating
 *       it, it is mandatory for you to use one.
 *       If no registry given, we will attempt to find a registry type matching
 *       the storage type.  
 *     - 'load_method' : const
 *       (optional) default is Xoxo_Storage::LOAD_METHOD_INSTANCE
 *       Two methods are right for this:
 *         - Xoxo_Storage::LOAD_METHOD_INSTANCE
 *           Normal method, this will only copy the loaded object reference, as
 *           node module will do with node_load(). This means every modification
 *           on a loaded object will be reflected to wherever this object has
 *           been loaded.
 *         - Xoxo_Storage::LOAD_METHOD_CLONE
 *           This method forces the load method to use the clone() function on
 *           static loaded instance each time you request it. This ensure that
 *           statically cached instances won't be modified (so any other module
 *           that asks to load an instance will have a fresh clean one).
 *     - 'class' : string
 *       (optional) default is 'Xoxo_Storage'
 *       Because you might want to add some behaviors depending on your object
 *       type, you'll then need to override the Xoxo_Storage by creating
 *       your own class that extends this one.
 *       You should always do that, for at least one good reason which is the
 *       Xoxo_Storage::exportCode() and Xoxo_Storage::importCode()
 *       functions are not implemented in default storage, and will throw an
 *       exception if you try to run them.
 *     - 'exportable' : boolean
 *       (optional) default is FALSE
 *       Tells if the objects are exportable or not. If TRUE, the custom storage
 *       must implement the exportCode() and importCode() methods.
 *     - 'target' string
 *       (mandatory)
 *       This is the classname of objects your module exports. If you use this
 *       facilities you must define at least your own objects that override
 *       either OoxExportable or OoxObject classes.
 *     - 'path' : string
 *       (optional) Give Drupal an different path for admin pages. This allows
 *       you to place your own objects administration screen wherever you want.
 *       Specifying this variable ensures that the generated URL's within those
 *       administration screens will hit the right path.
 *       This will only alter pathes, so if you are willing to use another title
 *       for any of the generated path, use the hook_menu_alter() to change it.
 */
function hook_xoxo_storage() {
  $items = array();
  // Non exportable data type. Ideal for site data. No UI will be present, but
  // the storage will easily be able to handle a really large amount of them.
  // UI, defaults, import and export features will be disable for this type, you
  // will only have access to the storage using:
  // <code>
  //   $storage = oox_storage_get('my_object_type');
  // </code>
  // This particular one will need that you defined the 'my_object_type'
  // registry using the hook_oox_registry().
  // Any objects from this registry will be CRUD-able using this storage.
  $items['my_object_type'] = array(
    'name' => "My simple object type",
    'table' => 'mymodule_my_object_type',
  );
  // Exportable data type without its own storage.
  // This object will be exportable using only the features module.
  // This sample also demonstrate that you can use a different registry to spwan
  // the storage supported objects.
  $items['my_exportable_type'] = array(
    'name' => "My exportable type",
    'table' => 'mymodule_my_exportable_type',
    'registry_type' => 'mymodule_my_registry_exportable_type',
    'exportable' => TRUE,
  );
  // Exportable data type with its own storage.
  // It will exportable using the features module, but also manually through
  // the menu under Administer > Build > My Full Type. You will also be able
  // to manage objects throught this UI.
  // This one implements the registry key, which will tell the storage to use
  // another registry than defined. It will also force the supported item types
  // that it can save from this registry.
  $items['my_full_type'] = array(
    'name' => "My Full Type",
    'administrable' => TRUE,
    // Next one can be omited.
    'table' => 'mymodule_object_type',
    // This is the default, can be omited.
    'load_method' => Xoxo_Storage::LOAD_METHOD_INSTANCE,
    'class' => 'MyObjectTypeStorage',
    // Sample of defining a subset of types given by the registry.
    'registry_type' => 'mymodule_my_registry_exportable_type',
    // Never leave the file empty, it is useless (it won't affect performances)
    // but just don't do it until you really need to specify a file.
    'file' => '',
    'target' => 'MyObjectType',
    'exportable' => FALSE,
    // Give another path for administration screens.
    'path' => 'admin/my_module/objects',
  );
  return $items;
}

/**
 * Return default instances defined by a custom module. This will allow site
 * administrator to use some objects without using the database.
 * 
 * This will also be the implemented hook when using the features module in
 * order to export site configuration.
 * 
 * @return array
 *   Array of default Xoxo_Object_Exportable instances.
 */
function hook_OBJECT_TYPE_defaults() {
  $items = array();
  $items[] = new OoxExportable('one', "My first instance");
  $items[] = new OoxExportable('two', "My second instance");
  $items[] = new OoxExportable('three', "My third instance");
  return $items;
}
